/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <stdio.h>
#include <varargs.h>

#include "misc.h"
#include "ix86_state.h"

static void (*efn)(FILE *stream, cpu_t *state) = NULL;


/***********************************************************************
 ***********************************************************************/
void
_error(char *file, char *func, int line, char *fmt, ...)
{
	va_list v;
	va_start(v);
  
	fprintf(stderr, "Run-time error in %s{%s,%d}:\n\t", func, file, line);
	vfprintf(stderr, fmt, v);
	fprintf(stderr, "\n");
	if (efn)
		(*efn)(stderr, NULL);
	abort();
	
	va_end(v);
}/*** end _error() ***/



/***********************************************************************
 ***********************************************************************/
void
RegisterErrorFn(void (*efn_arg)(FILE *stream, cpu_t *state))
{
	efn = efn_arg;

}/*** end RegisterErrorFn() ***/



/***********************************************************************
 ***********************************************************************/
void
_fatal(char *file, char *func, int line, char *fmt, ...)
{
	va_list v;
	va_start(v);

	fprintf(stderr, "Fatal error in %s{%s,%d}:\n\t", func, file, line);
	vfprintf(stderr, fmt, v);
	fprintf(stderr, "\n");
	
	va_end(v);
	
	exit(1);

}/*** end _fatal() ***/




/***********************************************************************
 ***********************************************************************/
void
_panic(char *file, char *func, int line, char *fmt, ...)
{
	va_list v;
	va_start(v);

	fprintf(stderr, "%s{%s,%d}:panic: ", func, file, line);
	vfprintf(stderr, fmt, v);
	fprintf(stderr, "\n");
	exit(1);

	va_end(v);

}/*** end _panid() ***/




/***********************************************************************
 ***********************************************************************/
void
_warn(char *file, char *func, int line, char *fmt, ...)
{
	va_list v;
	va_start(v);
  
	fprintf(stderr, "Warning in %s{%s,%d}:\n\t", func, file, line);
	vfprintf(stderr, fmt, v);
	fprintf(stderr, "\n");

	va_end(v);

}/*** end warn() ***/




/***********************************************************************
 ***********************************************************************/
void
_my_assert(char *file, char *func, int line, char *ex)
{
	fprintf(stderr, "Assert `%s' failed @ %s{%s,%d}\n", ex, func, file, line);
	abort();

}/*** end _my_assert() ***/





/***********************************************************************
   USAGE()
 ***********************************************************************/
void
Usage(int argc, char **argv)
{
	fprintf(stderr, "Usage: %s [-options] exec args...\n", argv[0]);

}/*** end Usage() ***/


/***********************************************************************
  signal handler routines used only to prevent core dumps if something
  bad happens.
 ***********************************************************************/
void
HandleSIGQUIT(long sigtype)
{  fprintf(stderr, "\ncaught SIGQUIT!\n\n");  exit(1); }

void
HandleSIGILL(long sigtype)
{  fprintf(stderr, "\ncaught SIGILL!\n\n");   exit(1); }

void
HandleSIGABRT(long sigtype)
{  fprintf(stderr, "\ncaught SIGABRT!\n\n");  exit(1); }

void
HandleSIGEMT(long sigtype)
{  fprintf(stderr, "\ncaught SIGEMT!\n\n");   exit(1); }

void
HandleSIGFPE(long sigtype)
{  fprintf(stderr, "\ncaught SIGFPE!\n\n");   exit(1); }

void
HandleSIGBUS(long sigtype)
{  fprintf(stderr, "\ncaught SIGBUS!\n\n");   exit(1); }

void
HandleSIGSEGV(long sigtype)
{  fprintf(stderr, "\ncaught SIGSEGV!\n\n");  exit(1); }

void
HandleSIGSYS(long sigtype)
{  fprintf(stderr, "\ncaught SIGSYS!\n\n");   exit(1); }

void
HandleSIGLOST(long sigtype)
{  fprintf(stderr, "\ncaught SIGLOST!\n\n");  exit(1); }
