/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#include <math.h>

#include "helper.h"
#include "ix86_state.h"
#include "operands.h"
#include "errorRoutines.h"

#define CF                (state->eflags.cf)
#define OF                (state->eflags.of)
#define ZF                (state->eflags.zf)
#define PF                (state->eflags.pf)
#define SF                (state->eflags.sf)
#define DF                (state->eflags.df)
#define AF                (state->eflags.af)
#define TF                (state->eflags.tf)
#define IF_               (state->eflags.if_)
#define NT                (state->eflags.nt)
#define TS                (state->cr0.ts)

#ifdef MICRO_OPS
/* note no semicolon following the macro to force the use of a semicolon */
#define MOPP micro_ops_this_cycle++
#define MOMM micro_ops_this_cycle--
#else
#define MOPP 
#define MOMM
#endif

/* This array defines a look-up table for the even parity-ness
   of an 8bit quantity, for optimal assignment of the parity bit
   in the EFLAGS register -- the 9-bit quantity of operand and parity bit
   produce an odd-parity quantity.
  */
static int parity_lookup[256] = {
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
	1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
};

static char xor_0x3_tab[] = { 0, 1, 1, 0 };



/****************************************************************
 not yet implemented rotate routines
 ****************************************************************/
Bit8u
rr_b(Bit8u op,  Bit8u bop, int ror, cpu_t *state)
{  assert(0); }

Bit16u
rr_w(Bit16u op, Bit8u bop, int ror, cpu_t *state)
{  assert(0); }

Bit32u
rr_d(Bit32u op, Bit8u bop, int ror, cpu_t *state)
{  assert(0); }

Bit8u
rl_b(Bit8u op,  Bit8u bop, int rol, cpu_t *state)
{  assert(0); }

Bit16u
rl_w(Bit16u op, Bit8u bop, int rol, cpu_t *state)
{  assert(0); }


/****************************************************************
 ****************************************************************/
inline
Bit32u
rl_d(Bit32u op, Bit8u bop, int rol, cpu_t *state)
{
	int     count;
	Boolean oldCF;
	Bit32u  shifted_out = 0,
	        res,
	        mask_d = 0;

	count = bop & 0x1f;
	/* assumes that the << operator shifts zero into low order bits */
	res = (op << count);

	if(count) {
		oldCF = (CF);
		assert(oldCF == 0 || oldCF == 1);

		(CF) = ((op & (0x80000000 >> (count-1))) != 0);
		mask_d = ~(   (0x80000000 >> (count-1)) - 1     );
		shifted_out = op & mask_d;
		shifted_out = shifted_out >> (32-count);

		/* sanity check on special case */
		assert(count || (shifted_out == 0));

		if(rol)
			res = res | shifted_out;
		else
			res = res | (shifted_out>>1) | (oldCF << count);

		if(count == 1)
			(OF) = ((CF) != ((op & 0x80000000) != 0));
	}
	
	if(verbose) {
		fprintf(stderr,"rl_d(): op = 0x%08lx, bop = 0x%02x, rol = %d, mask = 0x%08lx\n",
			op,
			bop,
			rol,
			mask_d);
		fprintf(stderr,"        shifted_out = 0x%08lx, res = 0x%08lx, cf = %s, of = %s\n",
			shifted_out,
			res,
			(CF) ? "TRUE" : "FALSE",
			(OF) ? "TRUE" : "FALSE");
	}

	return res;
}



/****************************************************************
 ****************************************************************/
Bit16u
shrd_w(Bit16u wop1, Bit16u wop2, Bit8u bop, cpu_t *state)
{
	assert(0);
	return 0;
}


/****************************************************************
 ****************************************************************/
Bit32u
shrd_d(Bit32u dop1, Bit32u dop2, Bit8u bop, cpu_t *state)
{
	int     shiftAmt;
	Bit64u  sources;
	Bit32u  dres;

	sources = ((Bit64u)dop2 <<32) | dop1;
	
	shiftAmt = bop % 32;
	
	if (shiftAmt == 0)
		return dop1;

	/* check to see if shift ammount is larger than the operands */
	/*
	   if (shiftAmt > 32)
	   return dop1;
	   */

	(CF) = (dop1 & (0x00000001 << (shiftAmt-1))) != 0;
	
	dres = (Bit32u)(sources >> shiftAmt);
	
	doubFlagsSZP(dres, state);

	if (verbose)
		fprintf(stderr,"shrd_d: dop2 = 0x%08lx, dop1 = 0x%08lx, bop = 0x%02x, res = 0x%08lx\n",
			dop2,dop1,bop,dres);

	return dres;
}


/****************************************************************
 ****************************************************************/
Bit16u
shld_w(Bit16u wop1, Bit16u wop2, Bit8u bop, cpu_t *state)
{
	assert(0);
	return 0;
}
/****************************************************************
 ****************************************************************/
Bit32u
shld_d(Bit32u dop1, Bit32u dop2, Bit8u bop, cpu_t *state)
{
	int     shiftAmt;
	Bit32u  dres;
	Bit64u  sources;

	shiftAmt = bop % 32;

	sources = ((Bit64u)dop1 <<32) | dop2;

	if (shiftAmt == 0)
		return dop1;

	/* check to see if shift ammount is larger than the operands */
	/*
	   if (shiftAmt > 32)
	   return dop1;
	   */

	(CF) = (dop1 & (0x80000000 >> (shiftAmt-1))) != 0;

	dres = (Bit32u)((sources << shiftAmt)>>32);
	
	doubFlagsSZP(dres, state);

	if (verbose)
		fprintf(stderr,"shld_d: dop1 = 0x%08lx, dop2 = 0x%08lx, bop = 0x%02x, res = 0x%08lx\n",
			dop1,dop2,bop,dres);
	
	return dres;
}

/****************************************************************
 ****************************************************************/
inline
Bit8u
shl_b(Bit8u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp,
              res;

	bop2 = bop2 & 0x1f;
	temp = bop2;
	res = op1;

	while(temp) {
		(CF) = ((res & 0x80) != 0);
		res = (res << 1) & 0xfe;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x80) != 0);
		(PF) = (parity_lookup[res & 0xff] == 1);
		(ZF) = (res == 0);

		if(bop2 == 1)
			(OF) = ((SF) != (CF));
	}

	if (verbose)
		fprintf(stderr,"shl_b: op1 = 0x%02x, bop = 0x%02x, res = 0x%02x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit8u
shr_b(Bit8u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp;
	Bit8u res;

	bop2 = bop2 & 0x1f;
	temp = bop2;
	res = op1;

	if(bop2 == 1)
	  (OF) = ((res & 0x80) != 0);

	while(temp) {
		(CF) = ((res & 0x01) != 0);
		res = (res >> 1) & 0x7f;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x80) != 0);
		(PF) = (parity_lookup[res & 0xff] == 1);
		(ZF) = (res == 0);
	}

	if (verbose)
		fprintf(stderr,"shr_b: op1 = 0x%02x, bop = 0x%02x, res = 0x%02x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit8u
sar_b(Bit8u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u bit,
              temp;
	Bit8u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	while(temp) {
		(CF) = ((res & 0x01) != 0);
		bit = res & 0x80;
		res = ((res >> 1) & 0x7f) | bit;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		if(bop2 == 1)
			(OF) = 0;

		(SF) = ((res&0x80) != 0);
		(PF) = (parity_lookup[res & 0xff] == 1);
		(ZF) = (res == 0);
	}

	if (verbose)
		fprintf(stderr,"sar_b: op1 = 0x%02x, bop = 0x%02x, res = 0x%02x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit16u
shl_w(Bit16u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp;
	Bit16u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	while(temp) {
		(CF) = ((res & 0x8000) != 0);
		res = (res << 1) & 0xfffe;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x8000) != 0);
		(PF) = (parity_lookup[res & 0x00ff] == 1);
		(ZF) = (res == 0);

		if(bop2 == 1)
			(OF) = ((SF) != (CF));
	}

	if (verbose)
		fprintf(stderr,"shl_w: op1 = 0x%04x, bop = 0x%02x, res = 0x%04x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit16u
shr_w(Bit16u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp;
	Bit16u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	if(temp == 1)
		(OF) = ((res & 0x8000) != 0);

	while(temp) {
		(CF) = ((res & 0x0001) != 0);
		res = (res >> 1) & 0x7fff;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x8000) != 0);
		(PF) = (parity_lookup[res & 0x00ff] == 1);
		(ZF) = (res == 0);
	}
	
	if (verbose)
		fprintf(stderr,"shr_w: op1 = 0x%04x, bop = 0x%02x, res = 0x%04x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit16u
sar_w(Bit16u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u bit,
              temp;
	Bit16u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	while(temp) {
		(CF) = ((res & 0x0001) != 0);
		bit = res & 0x8000;
		res = ((res >> 1) & 0x7fff) | bit;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		if(bop2 == 1)
			(OF) = 0;

		(SF) = ((res&0x8000) != 0);
		(PF) = (parity_lookup[res & 0x00ff] == 1);
		(ZF) = (res == 0);
	}

	if (verbose)
		fprintf(stderr,"sar_w: op1 = 0x%04x, bop = 0x%02x, res = 0x%04x\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit32u
shl_d(Bit32u op1,Bit8u bop2, cpu_t *state)
{
	Bit32u temp;
	Bit32u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	while(temp) {
		(CF) = ((res & 0x80000000) != 0);
		res = (res << 1) & 0xfffffffe;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x80000000) != 0);
		(PF) = (parity_lookup[res & 0x000000ff] == 1);
		(ZF) = (res == 0);

		if(bop2 == 1)
			(OF) = ((SF) != (CF));
	}

	if (verbose)
		fprintf(stderr,"shl_d: op1 = 0x%08lx, bop = 0x%02x, res = 0x%08lx\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit32u
shr_d(Bit32u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp;
	Bit32u res;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	if(bop2 == 1)
		(OF) = ((res & 0x80000000) != 0);

	while(temp) {
		(CF) = ((res & 0x00000001) != 0);
		res = (res >> 1) & 0x7fffffff;
		temp--;
	}

	/* flags change only if shift happens */
	if(bop2) {
		(SF) = ((res&0x80000000) != 0);
		(PF) = (parity_lookup[res & 0x000000ff] == 1);
		(ZF) = (res == 0);
	}

	if (verbose)
		fprintf(stderr,"shr_d: op1 = 0x%08lx, bop = 0x%02x, res = 0x%08lx\n",
			op1,bop2,res);

	return res;
}

/****************************************************************
 ****************************************************************/
inline
Bit32u
sar_d(Bit32u op1,Bit8u bop2, cpu_t *state)
{
	Bit8u temp;
	Bit32u res;
	Bit64u ll;

	ll = op1;
	if(op1 & 0x80000000)
		ll = ll | 0xffffffff00000000ULL;

	res = op1;

	bop2 = bop2 & 0x1f;
	temp = bop2;

	res = (Bit32u)(ll >> temp);

	/* flags change only if shift happens */
	if(bop2) {
		if(bop2 == 1)
			(OF) = 0;

		(SF) = ((res&0x80000000) != 0);
		(PF) = (parity_lookup[res & 0x000000ff] == 1);
		(ZF) = (res == 0);
	}

	if (verbose)
		fprintf(stderr,"sar_d: op1 = 0x%08lx, bop = 0x%02x, res = 0x%08lx\n",
			op1,bop2,res);

	return res;
}





/****************************************************************
 ****************************************************************/
inline
Bit32u
bsf(Bit32u dop, cpu_t *state)
{
	int  i;
	int  ret_val = 33;

	for (i = 0; i < 32; i++)
		if (dop & (1 << i)) {
			ret_val = i;
			break;
		}

	if(verbose)
		fprintf(stderr,"  bsf: dop1 = 0x%08lx, result = %d\n",dop,ret_val);

	return ret_val;
}

/****************************************************************
 ****************************************************************/
inline
Bit32u
bsr(Bit32u dop, cpu_t *state)
{
	int  i;
	int ret_val = -1;

	for (i = 31; i >= 0; i--)
		if (dop & (1 << i)) {
			ret_val = i;
			break;
		}

	if(verbose)
		fprintf(stderr,"  bsr: dop1 = 0x%08lx, result = %d\n",dop,ret_val);

	return ret_val;
}





/****************************************************************
 ****************************************************************/
inline
void
byteFlagsSZP(Bit8u res, cpu_t *state)
{
	SF = ((res&0x80) != 0);
	PF = (parity_lookup[res] == 1);
	ZF = (res == 0);
}

/****************************************************************
 ****************************************************************/
inline
void
wordFlagsSZP(Bit16u res, cpu_t *state)
{
	SF = ((res&0x8000) != 0);
	PF = (parity_lookup[res & 0x00ff] == 1);
	ZF = (res == 0);
}

/****************************************************************
 ****************************************************************/
inline
void
doubFlagsSZP(Bit32u res, cpu_t *state)
{
	SF = ((res&0x80000000) != 0);
	PF = (parity_lookup[res & 0x000000ff] == 1);
	ZF = (res == 0);
}

/****************************************************************
 ****************************************************************/
inline
void
byteFlagsAddOSZAP(Bit8u bop1, Bit8u bop2, Bit8u bres, cpu_t *state)
{
	Boolean bop2_b7, bop1_b7, bres_b7;
	Bit8u   cc;

	bop2_b7 = ((bop2 & 0x80) != 0);
	bop1_b7 = ((bop1 & 0x80) != 0);
	bres_b7 = ((bres & 0x80) != 0);

	cc = (bop1 & bop2) | ((~bres) & (bop1 | bop2));

	OF = xor_0x3_tab[(cc>>6)&0x03];
	SF = (bres_b7 != 0);
	ZF = (bres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[bres] == 1);
}

/****************************************************************
 ****************************************************************/
inline
void
wordFlagsAddOSZAP(Bit16u wop1, Bit16u wop2, Bit16u wres, cpu_t *state)
{
	Boolean wop2_b15, wop1_b15, wres_b15;
	Bit16u  cc;

	wop2_b15 = ((wop2 & 0x8000) != 0);
	wop1_b15 = ((wop1 & 0x8000) != 0);
	wres_b15 = ((wres & 0x8000) != 0);

	cc = (wop1 & wop2) | ((~wres) & (wop1 | wop2));

	OF = xor_0x3_tab[(cc>>14)&0x03];
	SF = (wres_b15 != 0);
	ZF = (wres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[wres & 0x00ff] == 1);
}

/****************************************************************
 ****************************************************************/
inline
void
doubFlagsAddOSZAP(Bit32u dop1, Bit32u dop2, Bit32u dres, cpu_t *state)
{
	Boolean dop2_b31, dop1_b31, dres_b31;
	Bit32u  cc;

	dop2_b31 = ((dop2 & 0x80000000) != 0);
	dop1_b31 = ((dop1 & 0x80000000) != 0);
	dres_b31 = ((dres & 0x80000000) != 0);

	cc = (dop1 & dop2) | ((~dres) & (dop1 | dop2));

	if(verbose)
		fprintf(stderr,"dop1 = 0x%08lx, dop2 = 0x%08lx, dres = 0x%08lx, cc = 0x%08lx\ndop2 & 0x80000000 = 0x%08lx,dop1 & 0x80000000 = 0x%08lx,dres & 0x80000000 = 0x%08lx\n",
			dop1,dop2,dres,cc,dop2 & 0x80000000,dop1 & 0x80000000,dres & 0x80000000);

	OF = xor_0x3_tab[(cc>>30)&0x03];
	SF = (dres_b31 != 0);
	ZF = (dres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[dres & 0x000000ff] == 1);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",	    (SF)?"SF":"",
			(ZF)?"ZF":"",	    (AF)?"AF":"",
			(PF)?"PF":"",	    (CF)?"CF":"");
}

/****************************************************************
 ****************************************************************/
inline
void
byteFlagsSubOSZAP(Bit8u bop1, Bit8u bop2, Bit8u bres, cpu_t *state)
{
	Boolean bop2_b7, bop1_b7, bres_b7;
	Bit8u   cc;

	cc = ((~bop1) & bop2) | (bres & ((~bop1) | bop2));

	if(verbose)
		fprintf(stderr,"bop1 = 0x%02x, bop2 = 0x%02x, bres = 0x%02x, cc = 0x%02x\n",
			bop1,bop2,bres,cc);

	bop2_b7 = ((bop2 & 0x80) != 0);
	bop1_b7 = ((bop1 & 0x80) != 0);
	bres_b7 = ((bres & 0x80) != 0);
	
	OF = xor_0x3_tab[(cc>>6)&0x03];
	SF = (bres_b7 != 0);
	ZF = (bres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[bres] == 1);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",	    (SF)?"SF":"",
			(ZF)?"ZF":"",	    (AF)?"AF":"",
			(PF)?"PF":"",	    (CF)?"CF":"");
}

/****************************************************************
 ****************************************************************/
inline
void
wordFlagsSubOSZAP(Bit16u wop1, Bit16u wop2, Bit16u wres, cpu_t *state)
{
	Boolean wop2_b15, wop1_b15, wres_b15;
	Bit16u  cc;

	cc = ((~wop1) & wop2) | (wres & ((~wop1) | wop2));

	wop2_b15 = ((wop2 & 0x00008000) != 0);
	wop1_b15 = ((wop1 & 0x00008000) != 0);
	wres_b15 = ((wres & 0x00008000) != 0);

	OF = xor_0x3_tab[(cc>>14)&0x03];
	SF = (wres_b15 != 0);
	ZF = (wres == 0);
	AF = ((cc & 0x0008) != 0);
	PF = (parity_lookup[wres & 0x00ff] == 1);
}


/****************************************************************
 ****************************************************************/
inline
void
doubFlagsSubOSZAP(Bit32u dop1, Bit32u dop2, Bit32u dres, cpu_t *state)
{
	Boolean dop2_b31, dop1_b31, dres_b31;
	Bit32u  cc;

	cc = ((~dop1) & dop2) | (dres & ((~dop1) | dop2));

	if(verbose)
		fprintf(stderr,"dop1 = 0x%08lx, dop2 = 0x%08lx, dres = 0x%08lx, cc = 0x%08lx\n",
			dop1,dop2,dres,cc);

	dop2_b31 = ((dop2 & 0x80000000) != 0);
	dop1_b31 = ((dop1 & 0x80000000) != 0);
	dres_b31 = ((dres & 0x80000000) != 0);

	OF = xor_0x3_tab[(cc>>30)&0x03];
	SF = (dres_b31 != 0);
	ZF = (dres == 0);
	AF = ((cc & 0x00000008) != 0);
	PF = (parity_lookup[dres & 0x000000ff] == 1);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",    (SF)?"SF":"",
			(ZF)?"ZF":"",    (AF)?"AF":"",
			(PF)?"PF":"",    (CF)?"CF":"");
}

/****************************************************************
 ****************************************************************/
inline
void
logicalByteFlags(Bit8u bres, cpu_t *state)
{
	OF = 0;
	CF = 0;
	byteFlagsSZP(bres, state);
}
/****************************************************************
 ****************************************************************/
inline
void
logicalWordFlags(Bit16u wres, cpu_t *state)
{
	OF = 0;
	CF = 0;
	wordFlagsSZP(wres, state);
}

/****************************************************************
 ****************************************************************/
inline
void
logicalDoubFlags(Bit32u dres, cpu_t *state)
{
	OF = 0;
	CF = 0;
	doubFlagsSZP(dres, state);
}


/****************************************************************
 ****************************************************************/
inline
void
standardAddByteFlags(Bit8u bop1, Bit8u bop2, Bit8u bres, cpu_t *state)
{
	Boolean bop2_b7, bop1_b7, bres_b7;
	Bit8u   cc;

	bop2_b7 = ((bop2 & 0x80) != 0);
	bop1_b7 = ((bop1 & 0x80) != 0);
	bres_b7 = ((bres & 0x80) != 0);

	cc = (bop1 & bop2) | ((~bres) & (bop1 | bop2));
	
	OF = xor_0x3_tab[(cc>>6)&0x03];
	SF = (bres_b7 != 0);
	ZF = (bres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[bres] == 1);
	CF = ((cc & 0x80) != 0);
}

/****************************************************************
 ****************************************************************/
inline
void
standardAddWordFlags(Bit16u wop1, Bit16u wop2, Bit16u wres, cpu_t *state)
{
	Boolean wop2_b15, wop1_b15, wres_b15;
	Bit16u  cc;

	wop2_b15 = ((wop2 & 0x8000) != 0);
	wop1_b15 = ((wop1 & 0x8000) != 0);
	wres_b15 = ((wres & 0x8000) != 0);
	
	cc = (wop1 & wop2) | ((~wres) & (wop1 | wop2));
	
	OF = xor_0x3_tab[(cc>>14)&0x03];
	SF = (wres_b15 != 0);
	ZF = (wres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[wres & 0x00ff] == 1);
	CF = ((cc & 0x8000) != 0);
}

/****************************************************************
 ****************************************************************/
inline
void
standardAddDoubFlags(Bit32u dop1, Bit32u dop2, Bit32u dres, cpu_t *state)
{
	Boolean dop2_b31, dop1_b31, dres_b31;
	Bit32u  cc;

	dop2_b31 = ((dop2 & 0x80000000) != 0);
	dop1_b31 = ((dop1 & 0x80000000) != 0);
	dres_b31 = ((dres & 0x80000000) != 0);

	cc = (dop1 & dop2) | ((~dres) & (dop1 | dop2));
	
	if(verbose)
		fprintf(stderr,"dop1 = 0x%08lx, dop2 = 0x%08lx, dres = 0x%08lx, cc = 0x%08lx\ndop2 & 0x80000000 = 0x%08lx,dop1 & 0x80000000 = 0x%08lx,dres & 0x80000000 = 0x%08lx\n",
			dop1,dop2,dres,cc,dop2 & 0x80000000,dop1 & 0x80000000,dres & 0x80000000);
	
	OF = xor_0x3_tab[(cc>>30)&0x03];
	SF = (dres_b31 != 0);
	ZF = (dres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[dres & 0x000000ff] == 1);
	CF = ((cc & 0x80000000) != 0);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",	    (SF)?"SF":"",
			(ZF)?"ZF":"",	    (AF)?"AF":"",
			(PF)?"PF":"",	    (CF)?"CF":"");
}


/****************************************************************
 ****************************************************************/
inline
void
standardSubByteFlags(Bit8u bop1, Bit8u bop2, Bit8u bres, cpu_t *state)
{
	Boolean bop2_b7, bop1_b7, bres_b7;
	Bit8u   cc;

	cc = ((~bop1) & bop2) | (bres & ((~bop1) | bop2));

	if(verbose)
		fprintf(stderr,"bop1 = 0x%02x, bop2 = 0x%02x, bres = 0x%02x, cc = 0x%02x\n",
			bop1,bop2,bres,cc);

	bop2_b7 = ((bop2 & 0x80) != 0);
	bop1_b7 = ((bop1 & 0x80) != 0);
	bres_b7 = ((bres & 0x80) != 0);

	OF = xor_0x3_tab[(cc>>6)&0x03];
	SF = (bres_b7 != 0);
	ZF = (bres == 0);
	AF = ((cc & 0x08) != 0);
	PF = (parity_lookup[bres] == 1);
	CF = ((cc & 0x80) != 0);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",	    (SF)?"SF":"",
			(ZF)?"ZF":"",	    (AF)?"AF":"",
			(PF)?"PF":"",	    (CF)?"CF":"");
}

/****************************************************************
 ****************************************************************/
inline
void
standardSubWordFlags(Bit16u wop1, Bit16u wop2, Bit16u wres, cpu_t *state)
{
	Boolean wop2_b15, wop1_b15, wres_b15;
	Bit16u  cc;

	cc = ((~wop1) & wop2) | (wres & ((~wop1) | wop2));
	
	wop2_b15 = ((wop2 & 0x00008000) != 0);
	wop1_b15 = ((wop1 & 0x00008000) != 0);
	wres_b15 = ((wres & 0x00008000) != 0);

	OF = xor_0x3_tab[(cc>>14)&0x03];
	SF = (wres_b15 != 0);
	ZF = (wres == 0);
	AF = ((cc & 0x0008) != 0);
	PF = (parity_lookup[wres & 0x00ff] == 1);
	CF = ((cc & 0x8000) != 0);
}

/****************************************************************
 ****************************************************************/
inline
void
standardSubDoubFlags(Bit32u dop1, Bit32u dop2, Bit32u dres, cpu_t *state)
{
	Boolean dop2_b31, dop1_b31, dres_b31;
	Bit32u  cc;

	cc = ((~dop1) & dop2) | (dres & ((~dop1) | dop2));

	if(verbose)
		fprintf(stderr,"dop1 = 0x%08lx, dop2 = 0x%08lx, dres = 0x%08lx, cc = 0x%08lx\n",
			dop1,dop2,dres,cc);

	dop2_b31 = ((dop2 & 0x80000000) != 0);
	dop1_b31 = ((dop1 & 0x80000000) != 0);
	dres_b31 = ((dres & 0x80000000) != 0);

	OF = xor_0x3_tab[(cc>>30)&0x03];
	SF = (dres_b31 != 0);
	ZF = (dres == 0);
	AF = ((cc & 0x00000008) != 0);
	PF = (parity_lookup[dres & 0x000000ff] == 1);
	CF = ((cc & 0x80000000) != 0);

	if(verbose)
		fprintf(stderr,"%s %s %s %s %s %s\n",
			(OF)?"OF":"",	    (SF)?"SF":"",
			(ZF)?"ZF":"",	    (AF)?"AF":"",
			(PF)?"PF":"",	    (CF)?"CF":"");
}


/***************************************************************************
 ***************************************************************************/
inline
void
stos(int size, iInst_t inst, cpu_t *state)
{
	Bit8u  bres = 0;
	Bit16u wres = 0;
	Bit32u dres = 0,
	       dadd;

	int    incDec = 0,
	       sign;
	Bit32u count;

	MOMM;

	assert(inst.addSize != WORD);

	switch (inst.repeat) {
	    case NO_REP:
		    count = 1;
		    break;

	    case REPE:
		MOPP; /* for compare */
		if ((count = getDoubOp(inst, 'C', state)) == 0)
			return;
		break;
	    default:     fatal("bad repeat enum member in stos.");
	}

	if(verbose)
		fprintf(stderr,"stos: count = 0x%08lx\n",count);

	if (DF)
		sign = -1;
	else
		sign = 1;

	switch(size) {
	    case 1: 
		    bres = getByteOp(inst, 'A', state);
		    break;
	    case 2:
		wres = getWordOp(inst, 'A', state);
		break;
	    case 4:
		dres = getDoubOp(inst, 'A', state);
		break;
	    default:
		    fatal("bad size in stos: %d",size);
	}

	do {
		/* 2 micro ops per iteration -- one for the store AL -> [edi] 
		   (done in MyMemoryAccess() and one for inc/dec of edi      */
		MOPP;  /* for inc/dec of edi */

		switch(size) {
		    case 1:
			storeByteOp(bres,inst,'d', state);
			incDec = 1;
			break;
		    case 2:
			storeWordOp(wres,inst,'d', state);
			incDec = 2;
			break;
		    case 4:
			storeDoubOp(dres,inst,'d', state);
			incDec = 4;
			break;
		}

		/* increment or decrement the edi index register */
		dadd = getDoubOp(inst,'D', state) + (incDec * sign);
		storeDoubOp(dadd,inst,'D', state);
		
		count--;
		
		if (inst.repeat == REPE) {
			/* decrement the count register */
			storeDoubOp(count, inst, 'C', state);
			MOPP; /* for subtraction */
		}
		MOPP; /* for compare to 0 */
	} while (count != 0);
}




/***************************************************************************
 ***************************************************************************/
inline
void
lods(int size, iInst_t inst, cpu_t *state)
{
	Bit8u  bres;
	Bit16u wres;
	Bit32u dres;

	int    incDec,
	       sign;
	Bit32u count;

	MOMM;

	assert(inst.addSize != WORD);

	switch (inst.repeat) {
	    case NO_REP:
		    count = 1;
		break;

	    case REPE:
		    MOPP; /* for compare */
		    if ((count = getDoubOp(inst, 'C', state)) == 0)
			    return;
		    break;
	    default:
		fatal("bad repeat enum member in stos.");
	}
	
	if(verbose)
		fprintf(stderr,"lods: count = 0x%08lx\n",count);

	if (DF)
		sign = -1;
	else
		sign = 1;

	do {
		MOPP; /* for inc/dec of edi */

		switch(size) {
		    case 1:
			    bres = getByteOp(inst, 'q', state);
			    storeByteOp(bres,inst,'A', state);
			    incDec = 1;
			    break;
		    case 2:
			    wres = getWordOp(inst, 'q', state);
			    storeWordOp(wres,inst,'A', state);
			    incDec = 2;
			    break;
		    case 4:
			    dres = getDoubOp(inst, 'q', state);
			    storeDoubOp(dres,inst,'A', state);
			    incDec = 4;
			    break;
		    default:
			    fatal("bad size in lods: %d",size);
		}

		/* increment or decrement the edi index register */
		dres = getDoubOp(inst,'Q', state) + (incDec * sign);
		storeDoubOp(dres,inst,'Q', state);

		count--;
		
		if(inst.repeat == REPE) {
			/* decrement the count register */
			storeDoubOp(count, inst, 'C', state);
			MOPP; /* for subtraction */
		}

		MOPP; /* for compare to 0 */
	} while (count != 0);
}


/***************************************************************************
 ***************************************************************************/
inline
void
cmps(int size, iInst_t inst, cpu_t *state)
{
	Bit8u  bop1, bop2, bres;
	Bit16u wop1, wop2, wres;
	Bit32u dop1, dop2, dres;
	int    sign,
	       incDec;
	Bit32u count;

	MOMM;

	assert(inst.addSize != WORD);

	switch (inst.repeat) {
	    case NO_REP:
		    count = 1;
		    break;

	    case REPE:
	    case REPNE:
		    MOPP; /* for compare */
		    if ((count = getDoubOp(inst, 'C', state)) == 0)
			    return;
		    break;
	    default:
		fatal("bad repeat enum member in scas.");
	}

	if (DF)
		sign = -1;
	else
		sign = 1;
	
	if(verbose)
		fprintf(stderr,"cmps: count = 0x%08lx\n",count);

	do {
		MOPP; /* for compare inside switch */
		MOPP; /* for inc/dec of edi */
		MOPP; /* for inc/dec of esi */

		switch(size) {

		    case 1:
			    bop1 = getByteOp(inst,'q', state);
			    bop2 = getByteOp(inst,'d', state);
			    if(verbose) fprintf(stderr,"[esi] = 0x%02x, [edi] = 0x%02x\n",bop1, bop2);
			    bres = bop1 - bop2;
			    standardSubByteFlags(bop1, bop2, bres, state);
			    incDec = 1;
			    break;
		    case 2:
			    wop1 = getWordOp(inst,'q', state);
			    wop2 = getWordOp(inst,'d', state);
			    if(verbose) fprintf(stderr,"[esi] = 0x%04x, [edi] = 0x%04x\n",wop1, wop2);
			    wres = wop1 - wop2;
			    standardSubWordFlags(wop1, wop2, wres, state);
			    incDec = 2;
			    break;
		    case 4:
			    dop1 = getDoubOp(inst,'q', state);
			    dop2 = getDoubOp(inst,'d', state);
			    if(verbose) fprintf(stderr,"[esi] = 0x%08lx, [edi] = 0x%08lx\n",dop1, dop2) ;
			    dres = dop1 - dop2;
			    standardSubDoubFlags(dop1, dop2, dres, state);
			    incDec = 4;
			    break;
		    default:
			fatal("bad size in cmps: %d",size);
		}

		/* increment or decrement the edi and esi index registers */
		dres = getDoubOp(inst,'D', state) + (incDec * sign);
		storeDoubOp(dres,inst,'D', state);
		dres = getDoubOp(inst,'Q', state) + (incDec * sign);
		storeDoubOp(dres,inst,'Q', state);
  
		count--;

		if(inst.repeat == REPE) {
			/* decrement the count register */
			MOPP; /* for subtraction */
			storeDoubOp(count, inst, 'C', state);
			if(!(ZF))
				return;
		}
		else
			if (inst.repeat == REPNE) {
				/* decrement the count register */
				MOPP; /* for subtraction */
				storeDoubOp(count, inst, 'C', state);
				if (ZF)
					return;
			}

		MOPP; /* for compare to 0 */
	} while(count != 0);
	
}



/***************************************************************************
 ***************************************************************************/
inline
void
scas(int size, iInst_t inst, cpu_t *state)
{
	Bit8u  bop1, bop2, bres;
	Bit16u wop1, wop2, wres;
	Bit32u dop1, dop2, dres;
	int    sign,
	       incDec;
	Bit32u count;

	assert(inst.addSize != WORD);

	MOMM;

	switch (inst.repeat) {
	case NO_REP: count = 1;
		break;

	case REPE:
	case REPNE:  MOPP; /* for compare */
		if ((count = getDoubOp(inst, 'C', state)) == 0)
                   return;
		break;
	default:     fatal("bad repeat enum member in scas.");
	}

	if(verbose)
		fprintf(stderr,"scas: count = 0x%08lx\n",count);

	if (DF)
		sign = -1;
	else
		sign = 1;

	do {
		MOPP; /* for compare inside switch */
		MOPP; /* for inc/dec of edi */

		switch(size) {
		case 1:
			bop1 = getByteOp(inst,'A', state);
			bop2 = getByteOp(inst,'d', state);
			if(verbose) fprintf(stderr,"al = 0x%02x, dd = 0x%02x\n",bop1, bop2);
			bres = bop1 - bop2;
			standardSubByteFlags(bop1, bop2, bres, state);
			incDec = 1;
			break;
		case 2:
			wop1 = getWordOp(inst,'A', state);
			wop2 = getWordOp(inst,'d', state);
			wres = wop1 - wop2;
			standardSubWordFlags(wop1, wop2, wres, state);
			incDec = 2;
			break;
		case 4:
			dop1 = getDoubOp(inst,'A', state);
			dop2 = getDoubOp(inst,'d', state);
			dres = dop1 - dop2;
			standardSubDoubFlags(dop1, dop2, dres, state);
			incDec = 4;
			break;
		default:
			fatal("bad size in scas: %d",size);
		}

		/* increment or decrement the edi index register */
		dres = getDoubOp(inst,'D', state) + (incDec * sign);
		storeDoubOp(dres,inst,'D', state);

		count--;
		
		if(inst.repeat == REPE) {
			/* decrement the count register */
			MOPP; /* for subtraction */
			storeDoubOp(count, inst, 'C', state);
			if(!(ZF))
				return;
		}
		else
			if (inst.repeat == REPNE) {
				/* decrement the count register */
				MOPP; /* for subtraction */
				storeDoubOp(count, inst, 'C', state);
				if (ZF)
					return;
			}
		
		MOPP; /* for compare to 0 */
	} while(count != 0);

}



/***************************************************************************
 ***************************************************************************/
inline
void
movs(int size, iInst_t inst, cpu_t *state)
{
	Bit8u  bres;
	Bit16u wres;
	Bit32u dres;

	int sign,
	    incDec;
	Bit32u count;

	assert(inst.addSize != WORD);

	MOMM;

	switch (inst.repeat) {
	    case NO_REP:
		    count = 1;
		    break;

	    case REPE:
		    MOPP;
		    if ((count = getDoubOp(inst, 'C', state)) == 0)
			    return;
		    break;
	    default:
		    fatal("bad r	epeat enum member in movs.");
	}

	if (DF)
		sign = -1;
	else
		sign = 1;

	if(verbose)
		fprintf(stderr,"movs: count = 0x%08lx\n",count);

	do {
		MOPP; /* for inc/dec of edi */
		MOPP; /* for inc/dec of esi */

		switch(size) {
		    case 1:
			    bres = getByteOp(inst,'q', state); /* [esi] */
			    storeByteOp(bres,inst,'d', state); /* [edi] */
			    incDec = 1;
			    break;
		    case 2:
			    wres = getWordOp(inst,'q', state);
			    storeWordOp(wres,inst,'d', state);
			    incDec = 2;
			    break;
		    case 4:
			    dres = getDoubOp(inst,'q', state);
			    storeDoubOp(dres,inst,'d', state);
			    incDec = 4;
			    break;
		    default:
			    fatal("bad size in movs: %d",size);
		}
		
		/* increment or decrement the edi index registers */
		dres = getDoubOp(inst,'D', state) + (incDec * sign);
		storeDoubOp(dres,inst,'D', state);
		
		/* increment or decrement the esi index registers */
		dres = getDoubOp(inst,'Q', state) + (incDec * sign);
		storeDoubOp(dres,inst,'Q', state);
		
		count--;
		
		if (inst.repeat == REPE) {
			/* decrement the count register */
			MOPP;
			storeDoubOp(count, inst, 'C', state);
		}

		MOPP;
	} while (count != 0);

}


/***************************************************************************
 ***************************************************************************/
Float64
round(Float64 fop,int round_mode)
{
	Float64 a,
	        c,
	        diffa,
	        diffc,
	        res;

	switch(round_mode) {
	case ROUND_NEAR_EVEN:
		a = floor(fop);
		c = ceil(fop);
		diffa = fabs(a - fop);
		diffc = fabs(c - fop);
		if(diffa < diffc)
			res = a; /* a is closer than c */
		else if(diffa > diffc)
			res = c; /* c is closer than a */
		else if (fmod(a,2.0)==0)
			/* a and c are equidistant -- pick the even one */
			res = a; /* a is even */
		else {
			assert(fmod(a,2.0) == 1);
			assert(fmod(c,2.0) == 0);
			res = c;  /* c must be even */
		}
		break;
    
	case ROUND_DOWN:
		res = floor(fop);
		break;
	case ROUND_UP:
		res = ceil(fop);
		break;
	case ROUND_CHOP:
		a = floor(fop);
		c = ceil(fop);
		if (fabs(a) <= fabs(c))
			res = a;
		else
			res = c;
		break;
	default:
		fatal("bad round control value in frndint.");
	}

	if(verbose)
		fprintf(stderr," round: op = %f, round_mode = %d, res = %f\n",fop, round_mode,res);

	return res;

}/*** end round() ***/



/***************************************************************************
 ***************************************************************************/
Float64
check_add64(Float64 ffop1, Float64 ffop2, Float64 ffres)
{
	Float64 ret_val = ffres;

	int inf1   = isinf(ffop1),
	    inf2   = isinf(ffop2),
	    infres = isinf(ffres);

	if(inf1 || inf2 || infres) {
		if(inf1 && inf2) {
			/* inf + inf */
			if((ffop1 > 0) && (ffop2 > 0)) {
				/* +inf + +inf = +inf */
				if((!infres) || (ffres <= 0))
					fprintf(stderr,"weird: add +inf + +inf != +inf!!!!!");
				
				ret_val = HUGE_VAL;
			}
			else if((ffop1 < 0) && (ffop2 < 0)) {
				/* -inf + -inf = -inf */
				if((!infres) || (ffres >= 0))
					fprintf(stderr,"weird: add -inf + -inf != -inf!!!!!");

				ret_val = -1 * HUGE_VAL;
			}
			else
				fatal("add invalid operation: +infinity + -infinity");
		} else if(inf1)
			ret_val = ffop1;
		else if(inf2)
			ret_val = ffop2;
		else {
			/* overflow???? */
			fatal("overflow in add???");
		}
	}

	return ret_val;
}
