/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef MEMORY_H
#define MEMORY_H

#include <stdio.h>
#include "misc.h"
#include "ix86_state.h"

/* VM-space details */
#if(0)
#define TEXT_BASE               0x08000080
#define DATA_TOP                0x7fffffff
#endif

/* page rounding macros */
#define PAGE_UP(addr)    ((((addr) + (PAGE_SIZE-1))/PAGE_SIZE)*PAGE_SIZE)

#define PAGE_SIZE               4096

enum MemCmd { Read, Write };

typedef void (*maft_t)(enum MemCmd cmd,		/* Read or Write */
		       unsigned long addr,		/* address to access */
		       unsigned char *p,		/* where to copy to/from */
		       int numBytes);		/* xfer length */

extern unsigned long minSP;	                /* min SP seen thus far */
extern unsigned long textEnd;			/* end of text */
extern unsigned long progEntry;			/* program entry point */
extern unsigned long breakPoint;		/* heap break point */
extern unsigned long baseBreak; 		/* initial heap break point */

#define MAXBUFSIZE 5000

extern inline Bit8u getIByte(unsigned long addr);

extern void MemoryAccess(enum MemCmd cmd,	/* Read or Write */
			 unsigned long addr,		/* address to access */
			 unsigned char *p,		/* where to copy to/from */
			 int numBytes);	/* size of the access */

extern void MemoryBlkAccess(enum MemCmd cmd,	/* Read or Write */
			    unsigned long addrOfBlk,	/* address to access */
			    unsigned long *p,		/* where to copy to/from */
			    int bSize);	/* size of the access */

extern unsigned long CopyToStr(maft_t maf, unsigned long addr, char *s);
extern unsigned long CopyFromStr(maft_t maf, char *s, unsigned long addr);
extern void          LoadProgram(int argc, char **argv, char **envp, cpu_t *state);
extern void          DumpMemory(unsigned long addr, unsigned long len);
extern void          MemStats(FILE *stream);

#endif /* MEMORY_H */
