/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

#ifndef OPERANDS_H
#define OPERANDS_H

#include "types.h"
#include "translate.h"
#include "memory.h"
#include "ix86_state.h"

extern Bit8u  getByteOp(iInst_t inst, char type, cpu_t *state);
extern Bit16u getWordOp(iInst_t inst, char type, cpu_t *state);
extern Bit32u getDoubOp(iInst_t inst, char type, cpu_t *state);
extern Bit64u getQuadOp(iInst_t inst, char type, cpu_t *state);
extern Bit32u getDoubAddr(iInst_t inst, char type, cpu_t *state);
extern void storeByteOp(Bit8u res, iInst_t inst, char type, cpu_t *state);
extern void storeWordOp(Bit16u res, iInst_t inst, char type, cpu_t *state);
extern void storeDoubOp(Bit32u res, iInst_t inst, char type, cpu_t *state);
extern void storeQuadOp(Bit64s res, iInst_t inst, char type, cpu_t *state);

extern void pop_fop(cpu_t *state);
extern Float64 fpu_stack(int depth, cpu_t *state);
extern void push_fop64(Float64 value, cpu_t *state);
extern void push_fop32(Float32 value, cpu_t *state);

extern Float32 getFloat32Op(iInst_t inst, char type, cpu_t *state);
extern Float64 getFloat64Op(iInst_t inst, char type, cpu_t *state);
extern void    storeFloat32Op(Float32 res, iInst_t inst, char type, cpu_t *state);
extern void    storeFloat64Op(Float64 res, iInst_t inst, char type, cpu_t *state);

extern void MyMemoryAccess(enum MemCmd cmd,              /* Read or Write */
		    unsigned long addr,           /* address to access */
		    unsigned char *p,             /* where to copy to/from */
		    int           numBytes);      /* size of the access */

#endif
