/* This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin, University of Wisconsin - Madison, Computer Sciences
 * Department as a part of the Multiscalar Research Project.
 *  
 * The SimpleScalar x86 port was written by Steve Bennett.
 * 
 * The tool suite is currently maintained by Doug Burger.
 * 
 * Copyright (C) 1994, 1995, 1996 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 */

/**************************************************************
 **************************************************************/
#include "translate.h"
#include "memory.h"
#include "ix86_state.h"
#include "misc.h"

/**************************************************************
 **************************************************************/
#define SIZE_OF_HASH 97

typedef struct hashEnt {
	unsigned long    raddress;
	iInst_t          inst;
	int              taddress;
	int              header;
	struct hashEnt   *next;
	struct hashEnt   *previous;
} hashEnt_t;

static hashEnt_t *hash[SIZE_OF_HASH];
static struct    hashEnt code[MAX_CODE_SIZE];
static int       nextTransIndex = 0;


enum indirect_type_enum {
	DIRECT = 0,
#define DEFFU(A,B)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE)
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP_NAME,ESCAPE,DECODERM,NAME) GROUP_NAME,
#include "ix86.def"
#undef DEFGROUP
#undef DEFLINK
#undef DEFINST
#undef DEFFU
	BAD_GROUP_HIGH
};

static char *group_name[BAD_GROUP_HIGH] = {
	"DIRECT",
#define DEFFU(A,B)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE)
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP_NAME,ESCAPE,DECODERM,NAME) #NAME,
#include "ix86.def"
#undef DEFGROUP
#undef DEFLINK
#undef DEFINST
#undef DEFFU
};
static int escape[BAD_GROUP_HIGH];
static int decoderm[BAD_GROUP_HIGH];

struct info_struct {
	enum indirect_type_enum indirect;
	enum Opcode opcode;
};
#define SIZE_OF_LARGEST_GROUP  2048
struct info_struct decode[BAD_GROUP_HIGH][SIZE_OF_LARGEST_GROUP];

static char *op2regb[] = { "AL",  "CL",  "DL",  "BL", 
			   "AH",  "CH",  "DH",  "BH"  };
static char *op2regw[] = { "AX",  "CX",  "DX",  "BX",
			   "SP",  "BP",  "SI",  "DI"  };
static char *op2regd[] = { "EAX", "ECX", "EDX", "EBX",
			   "ESP", "EBP", "ESI", "EDI" };



/* enum Opcode -> boolean indicating if a modrm byte is used by this opcode */
int modrmPresent[OP_MAX] = {
	0, /*OP_NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) MODRM,
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFFU
#undef DEFINST
#undef DEFLINK
};

/* enum Opcode ->  type of immediate field */
char op2immediate[OP_MAX] = {
	'*', /*OP_NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) IMMLEN,
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFFU
#undef DEFINST
#undef DEFLINK
};

/* enum Opcode -> opcode operand format */
char *op2format[OP_MAX] = {
	"XXX", /* NA */
#define DEFFU(FU,DESC)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) OPFORM,
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFFU
#undef DEFINST
#undef DEFLINK
};

typedef struct {
	Bit16u mod;
	Bit16u reg;
        Bit16u rm;
} modrm_decode_t;

static modrm_decode_t modrmtab[] = {
	{0,0,0}, {0,0,1}, {0,0,2}, {0,0,3}, /* 0 to 3 */
	{0,0,4}, {0,0,5}, {0,0,6}, {0,0,7}, /* 4 to 7 */
	{0,1,0}, {0,1,1}, {0,1,2}, {0,1,3}, /* 8 to 11 */
	{0,1,4}, {0,1,5}, {0,1,6}, {0,1,7}, /* 12 to 	15 */
	{0,2,0}, {0,2,1}, {0,2,2}, {0,2,3}, /* 16 to 19 */
	{0,2,4}, {0,2,5}, {0,2,6}, {0,2,7}, /* 20 to 23 */
	{0,3,0}, {0,3,1}, {0,3,2}, {0,3,3}, /* 24 to 27 */
	{0,3,4}, {0,3,5}, {0,3,6}, {0,3,7}, /* 28 to 31 */
	{0,4,0}, {0,4,1}, {0,4,2}, {0,4,3}, /* 32 to 35 */
	{0,4,4}, {0,4,5}, {0,4,6}, {0,4,7}, /* 36 to 39 */
	{0,5,0}, {0,5,1}, {0,5,2}, {0,5,3}, /* 40 to 43 */
	{0,5,4}, {0,5,5}, {0,5,6}, {0,5,7}, /* 44 to 47 */
	{0,6,0}, {0,6,1}, {0,6,2}, {0,6,3}, /* 48 to 51 */
	{0,6,4}, {0,6,5}, {0,6,6}, {0,6,7}, /* 52 to 55 */
	{0,7,0}, {0,7,1}, {0,7,2}, {0,7,3}, /* 56 to 59 */
	{0,7,4}, {0,7,5}, {0,7,6}, {0,7,7}, /* 60 to 63 */
	{1,0,0}, {1,0,1}, {1,0,2}, {1,0,3}, /* 64 to 67 */
	{1,0,4}, {1,0,5}, {1,0,6}, {1,0,7}, /* 68 to 71 */
	{1,1,0}, {1,1,1}, {1,1,2}, {1,1,3}, /* 72 to 75 */
	{1,1,4}, {1,1,5}, {1,1,6}, {1,1,7}, /* 76 to 79 */
	{1,2,0}, {1,2,1}, {1,2,2}, {1,2,3}, /* 80 to 83 */
	{1,2,4}, {1,2,5}, {1,2,6}, {1,2,7}, /* 84 to 87 */
	{1,3,0}, {1,3,1}, {1,3,2}, {1,3,3}, /* 88 to 91 */
	{1,3,4}, {1,3,5}, {1,3,6}, {1,3,7}, /* 92 to 95 */
	{1,4,0}, {1,4,1}, {1,4,2}, {1,4,3}, /* 96 to 99 */
	{1,4,4}, {1,4,5}, {1,4,6}, {1,4,7}, /* 100 to 103 */
	{1,5,0}, {1,5,1}, {1,5,2}, {1,5,3}, /* 104 to 107 */
	{1,5,4}, {1,5,5}, {1,5,6}, {1,5,7}, /* 108 to 111 */
	{1,6,0}, {1,6,1}, {1,6,2}, {1,6,3}, /* 112 to 115 */
	{1,6,4}, {1,6,5}, {1,6,6}, {1,6,7}, /* 116 to 119 */
	{1,7,0}, {1,7,1}, {1,7,2}, {1,7,3}, /* 120 to 123 */
	{1,7,4}, {1,7,5}, {1,7,6}, {1,7,7}, /* 124 to 127 */
	{2,0,0}, {2,0,1}, {2,0,2}, {2,0,3}, /* 128 to 131 */
	{2,0,4}, {2,0,5}, {2,0,6}, {2,0,7}, /* 132 to 135 */
	{2,1,0}, {2,1,1}, {2,1,2}, {2,1,3}, /* 136 to 139 */
	{2,1,4}, {2,1,5}, {2,1,6}, {2,1,7}, /* 140 to 143 */
	{2,2,0}, {2,2,1}, {2,2,2}, {2,2,3}, /* 144 to 147 */
	{2,2,4}, {2,2,5}, {2,2,6}, {2,2,7}, /* 148 to 151 */
	{2,3,0}, {2,3,1}, {2,3,2}, {2,3,3}, /* 152 to 155 */
	{2,3,4}, {2,3,5}, {2,3,6}, {2,3,7}, /* 156 to 159 */
	{2,4,0}, {2,4,1}, {2,4,2}, {2,4,3}, /* 160 to 163 */
	{2,4,4}, {2,4,5}, {2,4,6}, {2,4,7}, /* 164 to 167 */
	{2,5,0}, {2,5,1}, {2,5,2}, {2,5,3}, /* 168 to 171 */
	{2,5,4}, {2,5,5}, {2,5,6}, {2,5,7}, /* 172 to 175 */
	{2,6,0}, {2,6,1}, {2,6,2}, {2,6,3}, /* 176 to 179 */
	{2,6,4}, {2,6,5}, {2,6,6}, {2,6,7}, /* 180 to 183 */
	{2,7,0}, {2,7,1}, {2,7,2}, {2,7,3}, /* 184 to 187 */
	{2,7,4}, {2,7,5}, {2,7,6}, {2,7,7}, /* 188 to 191 */
	{3,0,0}, {3,0,1}, {3,0,2}, {3,0,3}, /* 192 to 195 */
	{3,0,4}, {3,0,5}, {3,0,6}, {3,0,7}, /* 196 to 199 */
	{3,1,0}, {3,1,1}, {3,1,2}, {3,1,3}, /* 200 to 203 */
	{3,1,4}, {3,1,5}, {3,1,6}, {3,1,7}, /* 204 to 207 */
	{3,2,0}, {3,2,1}, {3,2,2}, {3,2,3}, /* 208 to 211 */
	{3,2,4}, {3,2,5}, {3,2,6}, {3,2,7}, /* 212 to 215 */
	{3,3,0}, {3,3,1}, {3,3,2}, {3,3,3}, /* 216 to 219 */
	{3,3,4}, {3,3,5}, {3,3,6}, {3,3,7}, /* 220 to 223 */
	{3,4,0}, {3,4,1}, {3,4,2}, {3,4,3}, /* 224 to 227 */
	{3,4,4}, {3,4,5}, {3,4,6}, {3,4,7}, /* 228 to 231 */
	{3,5,0}, {3,5,1}, {3,5,2}, {3,5,3}, /* 232 to 235 */
	{3,5,4}, {3,5,5}, {3,5,6}, {3,5,7}, /* 236 to 239 */
	{3,6,0}, {3,6,1}, {3,6,2}, {3,6,3}, /* 240 to 243 */
	{3,6,4}, {3,6,5}, {3,6,6}, {3,6,7}, /* 244 to 247 */
	{3,7,0}, {3,7,1}, {3,7,2}, {3,7,3}, /* 248 to 251 */
	{3,7,4}, {3,7,5}, {3,7,6}, {3,7,7}, /* 252 to 255 */
};

/**************************************************************
 **************************************************************/
static
inline
int
index(unsigned long key)
{
	return (key % SIZE_OF_HASH);
}


/**************************************************************
  returns the address of the translated instruction for
  this intstruction, if it is found. Otherwise, returns NULL
 **************************************************************/
static
inline
iInst_t *
searchHash(unsigned long pc)
{
	int            hashInd;
	struct hashEnt *temp;

	hashInd = index(pc);

	temp = hash[hashInd];

	while(temp) {
		if (temp->raddress == pc) {
			numDecodeCacheHits++;
			return(&(temp->inst));
		}
		temp = temp->next;
	}
	return NULL;
}


/**************************************************************
 **************************************************************/
void
InitDecoder(void)
{
	int i,j;

	/* initialize hash table for real -> intermediate address */
	for (i = 0; i<SIZE_OF_HASH; i++)
		hash[i] = NULL;

	for (i = 0; i < MAX_CODE_SIZE; i++) {
		code[i].next = NULL;
		code[i].previous = NULL;
		code[i].taddress = -1;
		code[i].raddress = 0;
		code[i].header = 0;
		code[i].inst.realPC = 0;
	}

	/* "zero" the decoder arrays */
	for (i = DIRECT; i < BAD_GROUP_HIGH; i++)
		for (j = 0; j < SIZE_OF_LARGEST_GROUP; j++) {
			decode[i][j].indirect = DIRECT;
			decode[i][j].opcode = OP_NA;
		}



/* read in the escape and decoderm information */
#define DEFFU(A,B)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE)
#define DEFLINK(MSK,FROM,TO)
#define DEFGROUP(GROUP_NAME,ESCAPE,DECODERM,NAME) escape[GROUP_NAME]=ESCAPE; decoderm[GROUP_NAME]=DECODERM;
#include "ix86.def"
#undef DEFGROUP
#undef DEFLINK
#undef DEFINST
#undef DEFFU

/* read in the indirect links -- pointers to the groups and 2-byte opcode tables */
#define DEFFU(A,B)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE)
#define DEFLINK(MSK,FROM,TO) decode[FROM][MSK].indirect = TO;
#define DEFGROUP(GROUP_NAME,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFLINK
#undef DEFINST
#undef DEFFU



/* read in the opcodes into the appropriate tables */
#define DEFFU(A,B)
#define DEFLINK(OP,FROM,TO)
#define DEFINST(OP,OPCODE,MSK,WHERE,MODRM,IMMLEN,NAME,OPFORM,RES,FLAGS,O1,O2,I1,I2,I3,EXPR,TYPE) \
  if (MSK != 0)				\
  {					\
    for (i = 0; i<SIZE_OF_LARGEST_GROUP; i++)		\
    {					\
      if((i & (~(unsigned long)MSK)) == OPCODE)	\
        decode[WHERE][i].opcode = OP;	\
    }					\
  }					\
  else					\
    decode[WHERE][OPCODE].opcode = OP;



#define DEFGROUP(GROUP_NAME,ESCAPE,DECODERM,NAME)
#include "ix86.def"
#undef DEFGROUP
#undef DEFLINK
#undef DEFINST
#undef DEFFU
}



/**************************************************************
 **************************************************************/
void
dumpOpcodeInfo(FILE *stream)
{
	int i;
	int j;

	FILE *silly;

	silly = fopen("opcode.dump","w");

	assert(silly);

	for (i = DIRECT+1; i < BAD_GROUP_HIGH; i++) {
		fprintf(silly,"******************************************************\n");
		fprintf(silly,"Table #%d: %s\n",i,group_name[i]);
		fprintf(silly,"index    neumonic      ops  indirect?  mod?\timmediate\n");
		fprintf(silly,"******************************************************\n");
		for (j = 0; j<SIZE_OF_LARGEST_GROUP; j++) {
			fprintf(silly,"0x%03x: %10.10s %8.8s %10.10s %4.4s\t%c\n",
				j,
				op2name[decode[i][j].opcode],
				op2format[decode[i][j].opcode],
				(decode[i][j].indirect != 0) ?
				group_name[decode[i][j].indirect]:
				"-",
				(modrmPresent[decode[i][j].opcode] != 0) ?
				"Yes":
				"No",
				op2immediate[decode[i][j].opcode]);
		}
	}
	fclose(silly);
}


#define SCALED_INDEX (inst.index != 4) ? fprintf(stream,"%s*%d",op2regd[inst.index],1<<inst.ss):(fprintf(stream,"(0)"))
#define BASE         fprintf(stream, "%s",op2regd[inst.base])
#define D32          fprintf(stream,"0x%08lx",inst.displacement)
#define D8           fprintf(stream,"0x%02x",(Bit8u)inst.displacement)
/***********************************************************************
 ***********************************************************************/
static
void
printSIB(iInst_t inst, FILE *stream)
{
	fprintf(stream,"[");
	switch(inst.mod) {
	    case 0:
		if (inst.base == 5) {
			/* disp 32 + scaled index */
			D32;
			fprintf(stream," + ");
			SCALED_INDEX;
		} else {
			/* base + scaled index */
			BASE;
			fprintf(stream," + ");
			SCALED_INDEX;
		}
		break;
	    case 1:
		/* base + scaled index + d8 */
		BASE;
		fprintf(stream," + ");
		SCALED_INDEX;
		fprintf(stream," + ");
		D8;
		break;
	    case 2:
		/* base + scaled index + d32 */
		BASE;
		fprintf(stream," + ");
		D32;
		fprintf(stream," + ");
		SCALED_INDEX;
		break;
	default:
		fatal("bad mod value in printsib()");
	}
	fprintf(stream,"]");
}


/***********************************************************************
 ***********************************************************************/
void
DisassembleInst(iInst_t inst, FILE *stream, Bit32u eip)
{
	enum Opcode op;
	char *s;

	op = inst.op;
	s = OP_FORMAT(op);

	fprintf(stream, "0x%08lx:\t",inst.realPC);
	switch(inst.repeat) {
	    case NO_REP:
		fprintf(stream, "%-12s %c  ", OP_NAME(op),(inst.opSize == WORD) ? 'w': 'l');
		break;
	    case REPE:
		fprintf(stream, "repe %-7s %c  ", OP_NAME(op),(inst.opSize == WORD) ? 'w': 'l');
		break;
	    case REPNE:
		fprintf(stream, "repne %-6s %c  ", OP_NAME(op),(inst.opSize == WORD) ? 'w': 'l');
		break;
	    default:
		assert(0);
	}

	while (*s) {
		switch (*s) {
		case 'r':
		case 'R':
			/* modr/m byte */
			switch(inst.mod) {
			case 0:
				if (inst.rm == 4)
					/* sib byte only */
					printSIB(inst,stream);
				else if (inst.rm == 5)
					/* disp32 only */
					D32;
				else /* normal case */
					fprintf(stream,"[%s]",op2regd[inst.rm]);
				break;
			case 1:
				if (inst.rm == 4)
					/* sib byte only */
					printSIB(inst,stream);
				else {
					/* normal case */
					fprintf(stream,"[");
					D8;
					fprintf(stream," + %s]",op2regd[inst.rm]);
				}
				break;
			case 2:
				if (inst.rm == 4)
					/* sib byte only */
					printSIB(inst,stream);
				else {
					/* normal case */
					fprintf(stream,"[");
					D32;
					fprintf(stream," + %s]",op2regd[inst.rm]);
				}
				break;
			case 3:
				fprintf(stream,"%s",op2regd[inst.rm]);
				break;
			}
			break;
		case 'i': fprintf(stream,"$0x%02x",(Bit8u)(inst.immediate&0xff));
			break;
		case 'x': fprintf(stream,"$0x%04x",(Bit16u)(inst.immediate&0x0000ffff));
			break;
		case 'M': fprintf(stream,"0x%08lx",inst.immediate);
			break;
		case 'I': fprintf(stream,"$0x%08lx",inst.immediate);
			break;
		case 'o': fprintf(stream,"0x%08lx",signExtend8_32((Bit8u)(inst.offset&0xff))+eip+inst.iSize);
			break;
		case 'O': fprintf(stream,"0x%08lx",inst.offset+eip+inst.iSize);
			break;
		case 'a': fprintf(stream,"AL");
			break;
		case 'A': fprintf(stream,"EAX");
			break;
		case '|': fprintf(stream,"ST(%d)",inst.opcodeByte&7);
			break;
		case '+': fprintf(stream,"%s",op2regb[(inst.opcodeByte&7)]);
			break;
		case '=': if (inst.opSize == WORD)
			fprintf(stream,"%s",op2regw[(inst.opcodeByte&7)]);
		else
			fprintf(stream,"%s",op2regd[(inst.opcodeByte&7)]);
			break;
		case 'g': fprintf(stream,"%s",op2regb[inst.rmreg]);
			break;
		case 'G': if (inst.opSize == WORD)
			fprintf(stream,"%s",op2regw[inst.rmreg]);
		else
			fprintf(stream,"%s",op2regd[inst.rmreg]);
			break;
		case 'l': 
			s++;
			while((*s!='\0') && (*s!=',')) {
				fputc(*s, stream);
				s++;
			}
			if(*s)
				fputc(',',stream);
			else
				s--;
			break;
		case 'p':
		case 'S': fputc(*s, stream);
			break;
		default:
			fputc(*s, stream);
		}
		s++;
	}
	fprintf(stream, "\n");
}
#undef SCALED_INDEX
#undef BASE
#undef D32
#undef D8


/**************************************************************
 **************************************************************/
static
inline
Bit8u
getGroupIndex(Bit8u pibyte, Bit8u modrmbyte)
{
	/*         opcode bit 4,                    2,1,0          modr/m opcode */
	return (((pibyte >> 4) & 1) << 6) | ((pibyte & 0x7) << 3) | ((modrmbyte >> 3) & 0x7);
}

/***********************************************************************
 ***********************************************************************/
static
inline
iInst_t *
getTransEnt(unsigned long pc)
{
	int    hashind,
	       oldHashInd;
	struct hashEnt *temp;

	nextTransIndex ++;
	if (nextTransIndex >= MAX_CODE_SIZE)
		nextTransIndex  = 0;

	code[nextTransIndex].raddress = pc;
	hashind = index(pc);

	oldHashInd = code[nextTransIndex].taddress;

	if (oldHashInd != -1) {
		/* it is currently in a linked list */
		if (code[nextTransIndex].header) {
			/* it is first in the list at this bucket */
			hash[oldHashInd] = code[nextTransIndex].next;
			if (hash[oldHashInd] != NULL)
				hash[oldHashInd]->header = 1;
		} else {
			code[nextTransIndex].previous->next = code[nextTransIndex].next;
			if (code[nextTransIndex].next != NULL)
				code[nextTransIndex].next->previous = code[nextTransIndex].previous;
		}
	}
	/* else it is currently unused */

	code[nextTransIndex].taddress = hashind;
	code[nextTransIndex].header   = 1;
	code[nextTransIndex].previous = NULL;

	temp = hash[hashind];
	code[nextTransIndex].next = temp;
	hash[hashind]   = &code[nextTransIndex];
	if(temp != NULL) {
		temp->header = 0;
		temp->previous  = &code[nextTransIndex];
	}

	return (&(code[nextTransIndex].inst));
}


/***********************************************************************
 ***********************************************************************/
#define GETBYTE(START_ADDR) getIByte(START_ADDR)
#define GETWORD(START_ADDR)				\
               ((getIByte(START_ADDR+1) << 8 ) |	\
                (getIByte(START_ADDR+0)      )   )
#define GETDOUB(START_ADDR)				\
               ((getIByte(START_ADDR+3) << 24) |	\
                (getIByte(START_ADDR+2) << 16) |	\
                (getIByte(START_ADDR+1) << 8 ) |	\
                (getIByte(START_ADDR+0)      )   )

iInst_t
getInst(unsigned long pc)
{
	Bit8u                    modrm,
	                         sib;
	Bit16u	                 pibyte,
	                         rmgrpOpcode;
	int                      size;
	enum indirect_type_enum  grpName;
	enum Opcode              op;
	iInst_t                  *retInst;

	/* see if we have already been here */
	if ((retInst = searchHash(pc)) != NULL)
		return *retInst;
	
	/*
	 * in effect, this call does the insertion of the instruction
	 * into the hash table
	 */
	retInst = getTransEnt(pc);

	/* set default values for instruction */
	retInst->opSize  = DOUBLE;
	retInst->addSize = DOUBLE;
	retInst->repeat  = NO_REP;
	retInst->segOver = NO_OVER;
	retInst->fwait   = FALSE;

	/* get the first byte of the instruction */
	pibyte = getIByte(pc);
	assert(pibyte < 256);
	size = 1;

	/* walk through all prefix bytes that the instruction might have */
	while ((pibyte == 0x26) ||
	       (pibyte == 0x2e) ||
	       (pibyte == 0x36) ||
	       (pibyte == 0x3e) ||
	       (pibyte == 0x64) ||
	       (pibyte == 0x65) ||
	       (pibyte == 0x66) ||
	       (pibyte == 0x67) ||
	       (pibyte == 0xf0) ||
	       (pibyte == 0xf2) ||
	       (pibyte == 0xf3)) {
		switch(pibyte) {
		    case 0x26:
			/* seg = es */
			retInst->segOver = ES_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
		    case 0x36:
			/* seg = ss */
			retInst->segOver = SS_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
		case 0x2e:
			/* seg = cs */
			retInst->segOver = CS_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
		case 0x3e:
			/* seg = ds */
			retInst->segOver = DS_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
		case 0x64:
			/* seg = fs */
			retInst->segOver = FS_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
		case 0x65:
			/* seg = gs */
			retInst->segOver = GS_OVER;
			info("got a segment override prefix:%x",pibyte);
			break;
			
		case 0x66:
			/* operand size */
			retInst->opSize  = WORD;
			break;
			
		case 0x67:
			/* address size */
			retInst->addSize  = WORD;
			info("got an address size prefix:%x",pibyte);
			break;
			
		case 0xf0:
			/* lock */
			info("got a lock prefix:%x\n",pibyte);
			break;

		case 0xf2:
			/* repne */
			retInst->repeat = REPNE;
			break;
		case 0xf3:
			/* rep/repe */
			retInst->repeat = REPE;
			break;
		case 0x9b:
			/* fwait prefix */
			retInst->fwait = TRUE;
			break;
			
		    default:
			fatal("bad switch/while");
		}

		pibyte = getIByte(pc+size);
		assert(pibyte < 256);
		size++;
	}

	assert(ONE_BYTE == 1);
	grpName = decode[1][pibyte].indirect;
	if(grpName == DIRECT) {
		op = decode[1][pibyte].opcode;
		if(modrmPresent[op]) {
			modrm = getIByte(pc+size);
			assert(modrm < 256);
			size++;
		}
	} else {
		if(escape[grpName]) {
			pibyte = getIByte(pc+size);
			assert(pibyte < 256);
			size++;
		}

		if(decoderm[grpName]) {
			modrm = getIByte(pc+size);
			assert(modrm < 256);
			size++;
			if (grpName == FP_GRP) {
				rmgrpOpcode = (((pibyte & 0x07) << 8) | modrm);

				/* for debugging purposes only */
				pibyte = (pibyte <<8) | modrm;
			} else
				rmgrpOpcode = getGroupIndex(pibyte, modrm);
			op = decode[grpName][rmgrpOpcode].opcode;
		} else {
			op = decode[grpName][pibyte].opcode;
			if(modrmPresent[op]) {
				modrm = getIByte(pc+size);
				assert(modrm < 256);
				size++;
			}
		}
	}

	retInst->opcodeByte = pibyte;
	retInst->op         = op;
	retInst->realPC     = pc;
	if(modrmPresent[op]) {
		retInst->mod   = modrmtab[modrm].mod;
		retInst->rm    = modrmtab[modrm].rm;
		retInst->rmreg = modrmtab[modrm].reg;

		if ((retInst->mod != 3) && (retInst->rm == 4)) {
			/* sib byte present */
			sib = getIByte(pc+size);
			assert(sib < 256);
			size++;
			retInst->ss    = modrmtab[sib].mod;
			retInst->index = modrmtab[sib].reg;
			retInst->base  = modrmtab[sib].rm;
			
			switch(retInst->mod) {
			    case 0:
				if(retInst->base == 5) {
					/* there is a displacement 32 indicated by the sib byte */
					retInst->displacement = (Bit32u)GETDOUB(pc+size);
					size += 4;
				}
				break;
			    case 1:
				retInst->displacement = (Bit32u)GETBYTE(pc+size);
				size ++;
				break;
			    case 2:
				retInst->displacement = (Bit32u)GETDOUB(pc+size);
				size += 4;
				break;
			    case 3:
			    default:
				assert(0);
				break;
			}
		} else
			switch(retInst->mod) {
			    case 0:
				    if(retInst->rm == 5) {
					    retInst->displacement = (Bit32u)GETDOUB(pc+size);
					    size += 4;
				    }
				    break;
			    case 1:
				    retInst->displacement = (Bit32u)GETBYTE(pc+size);
				    size++;
				    break;
			    case 2:
				    retInst->displacement = (Bit32u)GETDOUB(pc+size);
				    size += 4;
				    break;
			    case 3:
				    break;
			    default:
				    assert(0);
			}
	}


	switch (op2immediate[op]) {
	    case '_':
		break;

	    case 'o':
		retInst->offset = (Bit32u)GETBYTE(pc+size);
		size++;
		break;
	    case 'O':
		assert(retInst->addSize != WORD);
		retInst->offset = (Bit32u)GETDOUB(pc+size);
		size += 4;
		break;

	    case 'i':
		retInst->immediate = (Bit32u)GETBYTE(pc+size);
		size++;
		break;

	    case 'x':
		/* ret imm16 immediate */
		retInst->immediate = (Bit32u)GETWORD(pc+size);
		size += 2;
		break;
	    case 'I':
		if (retInst->opSize == WORD) {
			retInst->immediate = (Bit32u)GETWORD(pc+size);
			size += 2;
		} else {
			retInst->immediate = (Bit32u)GETDOUB(pc+size);
			size += 4;
		}
		break;

	    case 'M':
		assert(retInst->addSize != WORD);
		retInst->immediate = (Bit32u)GETDOUB(pc+size);
		size += 4;
		break;

	    case 'P':
		printf("don't know how to do P yet\n");
		size += 6;
		break;

	    default:
		fatal("bad immediate format %c at pc = 0x%08lx",op2immediate[op],pc);
	}

	assert(size <= 16);
	
	retInst->iSize = size;

	return *retInst;
}
#undef GETBYTE(START_ADDR)
#undef GETWORD(START_ADDR)
#undef GETDOUB(START_ADDR)
